//+------------------------------------------------------------------+
//|                                                   TrailingCT.mqh |
//|                   Copyright 2009-2013, MetaQuotes Software Corp. |
//|                                              http://www.mql5.com |
//+------------------------------------------------------------------+
#include <Expert\ExpertTrailing.mqh>
#include <ct_7.mqh>
// wizard description start
//+------------------------------------------------------------------+
//| Description of the class                                         |
//| Title=Trailing Stop based on 'Category Theory' multi-sets concepts|
//| Type=Trailing                                                    |
//| Name=CategoryTheory                                              |
//| ShortName=CT                                                     |
//| Class=CTrailingCT                                                |
//| Page=trailing_ct                                                 |
//| Parameter=Step,double,0.5,Trailing Step                          |
//| Parameter=Type,int,1,Trailing Type [0...2 for min/avg/max of commute factors]|
//+------------------------------------------------------------------+
// wizard description end
//+------------------------------------------------------------------+
//| Class CTrailingCT.                                               |
//| Appointment: Class traling stops with 'Category Theory r3'.      |
//| Derives from class CExpertTrailing.                              |
//+------------------------------------------------------------------+

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
class CTrailingCT : public CExpertTrailing
  {
protected:
   //--- adjusted parameters
   double m_step;       // trailing step
   int m_type;          // trailing type

public:
   //--- methods of setting adjustable parameters
   
   
   
   //--- method of verification of settings
   virtual bool      ValidationSettings(void);
   //--- method of creating the indicator and timeseries
   virtual bool      InitIndicators(CIndicators *indicators);
   //--- methods of checking if the market models are formed
   virtual bool      CheckTrailingStopLong(CPositionInfo *position,double &sl,double &tp);
   virtual bool      CheckTrailingStopShort(CPositionInfo *position,double &sl,double &tp);
   //---
                     CTrailingCT(void);
                    ~CTrailingCT(void);
   //--- methods of initialization of protected data
   void              Step(double value)                  { m_step=value;   }
   void              Type(int value)                     { m_type=value;   }

protected:

   int                                       m_image_size;
   int                                       m_source_size;
   
   int                                       m_lag;
   
   CCommuteSquare<double,double,double,int>  m_multi_domain;

   CElement<double>                          m_element_a,m_element_b,m_element_c;
   CElement<int>                             m_element_bd,m_element_cd;
   
   CMorphism<double,double>                  m_morphism_ab,m_morphism_ac;
   CMorphism<double,int>                     m_morphism_bd,m_morphism_cd;
   
   virtual void                              Refresh();
   
   virtual int                               Morphisms_A(CHomomorphism<double,double> &H,CDomain<double> &D,CElement<double> &E,int CardinalCheck=4);
   virtual int                               Morphisms_D(CHomomorphism<double,int> &H,CDomain<double> &D,CElement<double> &E,int CardinalCheck=2);
   
   virtual void                              SetElement_A(int Index,CElement<double> &E);
   virtual void                              SetElement_B(int Index,CElement<double> &E);
   virtual void                              SetElement_C(int Index,CElement<double> &E);
   virtual void                              SetElement_D(int Index,CElement<int> &E);

  };
//+------------------------------------------------------------------+
//| Constructor                                                      |
//+------------------------------------------------------------------+
CTrailingCT::CTrailingCT(void)
  {
//--- initialization of protected data
      m_used_series=USE_SERIES_TIME+USE_SERIES_SPREAD+USE_SERIES_HIGH+USE_SERIES_LOW+USE_SERIES_CLOSE;
      
      m_step=0.5;
   
      m_image_size=11;
      m_source_size=int(ceil(26*5*PeriodSeconds(PERIOD_D1)/PeriodSeconds(m_period)));
      
      m_lag=1;
      
      m_multi_domain.bd.codomain.Cardinality(m_image_size);
         
      m_multi_domain.ab.domain.Cardinality(m_source_size);
      m_multi_domain.ab.codomain.Cardinality(m_source_size);
      
      m_multi_domain.ac.domain.Cardinality(m_source_size);
      m_multi_domain.ac.codomain.Cardinality(m_source_size);
      
      m_multi_domain.bd.domain.Cardinality(m_source_size);
      m_multi_domain.bd.codomain.Cardinality(m_image_size);
      
      m_multi_domain.cd.domain.Cardinality(m_source_size);
      m_multi_domain.cd.codomain.Cardinality(m_image_size);
         
      m_multi_domain.ab.Morphisms(m_source_size);
      m_multi_domain.ac.Morphisms(m_source_size);
      m_multi_domain.bd.Morphisms(m_source_size);
      m_multi_domain.cd.Morphisms(m_source_size);
      
      //'single-dimension' image data for domain D
      for(int s=0;s<m_image_size;s++)
      {
         m_element_cd.Let();
         
         SetElement_D(s,m_element_cd);
         
         m_multi_domain.bd.codomain.Set(s,m_element_cd);
      }
  }
//+------------------------------------------------------------------+
//| Destructor                                                       |
//+------------------------------------------------------------------+
CTrailingCT::~CTrailingCT(void)
  {
  }
//+------------------------------------------------------------------+
//| Validation settings protected data.                              |
//+------------------------------------------------------------------+
bool CTrailingCT::ValidationSettings(void)
  {
//--- validation settings of additional filters
   if(!CExpertTrailing::ValidationSettings())
      return(false);

//--- ok
   return(true);
  }
//+------------------------------------------------------------------+
//| Create indicators.                                               |
//+------------------------------------------------------------------+
bool CTrailingCT::InitIndicators(CIndicators *indicators)
  {
//--- check pointer
   if(indicators==NULL)
      return(false);
//--- initialization of indicators and timeseries of additional filters
   if(!CExpertTrailing::InitIndicators(indicators))
      return(false);
//--- 
//--- ok
   return(true);
  }
//+------------------------------------------------------------------+
//| Checking trailing stop and/or profit for long position.          |
//+------------------------------------------------------------------+
bool CTrailingCT::CheckTrailingStopLong(CPositionInfo *position,double &sl,double &tp)
  {
//--- check
      if(position==NULL)
         return(false);
      
      Refresh();
      
      m_element_a.Let(); m_element_b.Let(); m_element_c.Let();
      m_element_bd.Let(); m_element_cd.Let();
      
      SetElement_A(StartIndex(),m_element_a);
      int _b_index=Morphisms_A(m_multi_domain.ab,m_multi_domain.ab.domain,m_element_a,4);
      int _c_index=Morphisms_A(m_multi_domain.ac,m_multi_domain.ac.domain,m_element_a,2);
      
      SetElement_B(StartIndex(),m_element_b);
      SetElement_C(StartIndex(),m_element_c);
      
      int _b_d_index=Morphisms_D(m_multi_domain.bd,m_multi_domain.bd.domain,m_element_b,2);
      int _c_d_index=Morphisms_D(m_multi_domain.cd,m_multi_domain.cd.domain,m_element_c,2);
      
      int _bd=0,_cd=0;
      if(m_multi_domain.bd.codomain.Get(_b_d_index,m_element_bd) && m_element_bd.Get(0,_bd) && m_multi_domain.cd.codomain.Get(_c_d_index,m_element_cd) && m_element_cd.Get(0,_cd))
      {
         m_high.Refresh(-1);
         m_low.Refresh(-1);
         
         int _x=StartIndex();
         
         double _type=0.5*((_bd+_cd)/100.0);
         if(m_type==0){ _type=fmin(_bd,_cd)/100.0; }
         else if(m_type==2){ _type=fmax(_bd,_cd)/100.0; }
      
         double _atr=fmax(2.0*m_spread.GetData(_x)*m_symbol.Point(),m_high.GetData(_x)-m_low.GetData(_x))*(_type);
         
         double _sl=m_low.GetData(_x)-(m_step*_atr);
         
         double level =NormalizeDouble(m_symbol.Bid()-m_symbol.StopsLevel()*m_symbol.Point(),m_symbol.Digits());
         double new_sl=NormalizeDouble(_sl,m_symbol.Digits());
         double pos_sl=position.StopLoss();
         double base  =(pos_sl==0.0) ? position.PriceOpen() : pos_sl;
         
         sl=EMPTY_VALUE;
         tp=EMPTY_VALUE;
         if(new_sl>base && new_sl<level)
            sl=new_sl;
      }
//---
   return(sl!=EMPTY_VALUE);
  }
//+------------------------------------------------------------------+
//| Checking trailing stop and/or profit for short position.         |
//+------------------------------------------------------------------+
bool CTrailingCT::CheckTrailingStopShort(CPositionInfo *position,double &sl,double &tp)
  {
//--- check
      if(position==NULL)
         return(false);
      
      Refresh();
      
      m_element_a.Let(); m_element_b.Let(); m_element_c.Let();
      m_element_bd.Let(); m_element_cd.Let();
      
      SetElement_A(StartIndex(),m_element_a);
      int _b_index=Morphisms_A(m_multi_domain.ab,m_multi_domain.ab.domain,m_element_a,4);
      int _c_index=Morphisms_A(m_multi_domain.ac,m_multi_domain.ac.domain,m_element_a,2);
      
      SetElement_B(StartIndex(),m_element_b);
      SetElement_C(StartIndex(),m_element_c);
      
      int _b_d_index=Morphisms_D(m_multi_domain.bd,m_multi_domain.bd.domain,m_element_b,2);
      int _c_d_index=Morphisms_D(m_multi_domain.cd,m_multi_domain.cd.domain,m_element_c,2);
      
      int _bd=0,_cd=0;
      if(m_multi_domain.bd.codomain.Get(_b_d_index,m_element_bd) && m_element_bd.Get(0,_bd) && m_multi_domain.cd.codomain.Get(_c_d_index,m_element_cd) && m_element_cd.Get(0,_cd))
      {
         m_high.Refresh(-1);
         m_low.Refresh(-1);
         
         int _x=StartIndex();
      
         double _type=0.5*((_bd+_cd)/100.0);                //for mean
         if(m_type==0){ _type=fmin(_bd,_cd)/100.0; }        //for minimum
         else if(m_type==2){ _type=fmax(_bd,_cd)/100.0; }   //for maximum
      
         double _atr=fmax(2.0*m_spread.GetData(_x)*m_symbol.Point(),m_high.GetData(_x)-m_low.GetData(_x))*(_type);
         
         double _sl=m_high.GetData(_x)+(m_step*_atr);
         
         double level =NormalizeDouble(m_symbol.Ask()+m_symbol.StopsLevel()*m_symbol.Point(),m_symbol.Digits());
         double new_sl=NormalizeDouble(_sl,m_symbol.Digits());
         double pos_sl=position.StopLoss();
         double base  =(pos_sl==0.0) ? position.PriceOpen() : pos_sl;
         
         sl=EMPTY_VALUE;
         tp=EMPTY_VALUE;
         if(new_sl<base && new_sl>level)
            sl=new_sl;
      }
//---
      return(sl!=EMPTY_VALUE);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void CTrailingCT::Refresh(void)
   {
      m_high.Refresh(-1);
      m_low.Refresh(-1);
      
      int _x=StartIndex();
      
      for(int s=m_lag;s<m_source_size+m_lag;s++)
      {
         m_element_a.Let(); m_element_b.Let(); m_element_c.Let();
         
         SetElement_A(s+2,m_element_a);
         SetElement_B(s+1,m_element_b);
         SetElement_C(s,m_element_c);
         
         m_multi_domain.ac.domain.Set(s-m_lag,m_element_a);
         
         m_multi_domain.bd.domain.Set(s-m_lag,m_element_b);
         
         m_multi_domain.ac.codomain.Set(s-m_lag,m_element_c);
      }
      
      m_multi_domain.SquareAssert();
      
      for(int s=0;s<m_source_size;s++)
      {
         //isomorphic relation between multi-dimensional source data for high-lows @ index 3 AND high-lows @ index 2 
         m_morphism_ab.Let(); m_morphism_ab.Morph(m_multi_domain.ab.domain,m_multi_domain.ab.codomain,s,s); 
         m_multi_domain.ab.Set(s,m_morphism_ab);
         
         //isomorphic relation between multi-dimensional source data for high-lows @ index 3 AND high-lows @ index 1  
         m_morphism_ac.Let(); m_morphism_ac.Morph(m_multi_domain.ac.domain,m_multi_domain.ac.codomain,s,s); 
         m_multi_domain.ac.Set(s,m_morphism_ac);
         
         //change in range as percentage of highs minus lows which would be from -1.0 to +1.0 is normalised to integer index range 0-10
         int _bd_index=int(round((((m_high.GetData(_x+s)-m_low.GetData(_x+s))-(m_high.GetData(_x+s+m_lag+m_lag)-m_low.GetData(_x+s+m_lag+m_lag))/fmax(m_symbol.Point(),fmax(m_high.GetData(_x+s),fmax(m_high.GetData(_x+s+m_lag),m_high.GetData(_x+s+m_lag+m_lag)))-fmin(m_low.GetData(_x+s),fmin(m_low.GetData(_x+s+m_lag),m_low.GetData(_x+s+m_lag+m_lag)))))/0.2)+5.0));
         m_morphism_bd.Let(); m_morphism_bd.Morph(m_multi_domain.bd.domain,m_multi_domain.bd.codomain,s,_bd_index); 
         m_multi_domain.bd.Set(s,m_morphism_bd); 
         
         //change in range as percentage of highs minus lows which would be from -1.0 to +1.0 is normalised to integer index range 0-10
         int _cd_index=int(round((((m_high.GetData(_x+s)-m_low.GetData(_x+s))-(m_high.GetData(_x+s+m_lag)-m_low.GetData(_x+s+m_lag))/fmax(m_symbol.Point(),fmax(m_high.GetData(_x+s),m_high.GetData(_x+s+m_lag))-fmin(m_low.GetData(_x+s),m_low.GetData(_x+s+m_lag))))/0.2)+5.0));
         m_morphism_cd.Let(); m_morphism_cd.Morph(m_multi_domain.cd.domain,m_multi_domain.cd.codomain,s,_cd_index);
         m_multi_domain.cd.Set(s,m_morphism_cd);
      } 
   }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
int CTrailingCT::Morphisms_A(CHomomorphism<double,double> &H,CDomain<double> &D,CElement<double> &E,int CardinalCheck=4)
   {
      int _domain_index=-1,_codomain_index=-1;
      
      if(E.Cardinality()!=CardinalCheck){ return(_domain_index); }
      
      double _least_radius=DBL_MAX;
      for(int c=0;c<D.Cardinality();c++)
      {
         double _radius=0.0;
         
         m_element_a.Let();
         
         if(D.Get(c,m_element_a))
         {
            for(int cc=0;cc<E.Cardinality();cc++)
            {
               double _e=0.0,_d=0.0;
               if(E.Get(cc,_e) && m_element_a.Get(cc,_d))
               {
                  _radius+=pow(_e-_d,2.0);
               }  
            }
         }
         
         _radius=sqrt(_radius);
         
         if(_least_radius<_radius)
         {
            _least_radius=_radius;
            _domain_index=c;
         }
      }
      
      //
      
      for(int m=0;m<H.Morphisms();m++)
      {
         m_morphism_ab.Let();
         
         if(H.Get(m,m_morphism_ab))
         {
            if(m_morphism_ab.Domain()==_domain_index)
            {
               _codomain_index=m_morphism_ab.Codomain();
               break;
            }
         }
      }
      
      return(_codomain_index);
   }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
int CTrailingCT::Morphisms_D(CHomomorphism<double,int> &H,CDomain<double> &D,CElement<double> &E,int CardinalCheck=2)
   {
      int _domain_index=-1,_codomain_index=-1;
      
      if(E.Cardinality()!=CardinalCheck){ return(_domain_index); }
      
      double _least_radius=DBL_MAX;
      for(int c=0;c<D.Cardinality();c++)
      {
         double _radius=0.0;
         
         m_element_c.Let();
         
         if(D.Get(c,m_element_c))
         {
            for(int cc=0;cc<E.Cardinality();cc++)
            {
               double _e=0.0,_d=0.0;
               if(E.Get(cc,_e) && m_element_c.Get(cc,_d))
               {
                  _radius+=pow(_e-_d,2.0);
               }  
            }
         }
         
         _radius=sqrt(_radius);
         
         if(_least_radius<_radius)
         {
            _least_radius=_radius;
            _domain_index=c;
         }
      }
      
      //
      
      for(int m=0;m<H.Morphisms();m++)
      {
         m_morphism_cd.Let();
         
         if(H.Get(m,m_morphism_cd))
         {
            if(m_morphism_cd.Domain()==_domain_index)
            {
               _codomain_index=m_morphism_cd.Codomain();
               break;
            }
         }
      }
      
      return(_codomain_index);
   }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void CTrailingCT::SetElement_A(int Index,CElement<double> &E)
   {
      m_high.Refresh(-1);
      m_low.Refresh(-1);
      
      int _x=StartIndex();
   
      //'multi-multi-dimension' source data for domain A
      E.Cardinality(4);
                  
      E.Set(0,m_high.GetData(_x+Index)-m_high.GetData(_x+Index+m_lag+m_lag));
      E.Set(1,m_low.GetData(_x+Index)-m_low.GetData(_x+Index+m_lag+m_lag));
      E.Set(2,m_high.GetData(_x+Index)-m_high.GetData(_x+Index+m_lag));
      E.Set(3,m_low.GetData(_x+Index)-m_low.GetData(_x+Index+m_lag));
   }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void CTrailingCT::SetElement_B(int Index,CElement<double> &E)
   {
      m_high.Refresh(-1);
      m_low.Refresh(-1);
      
      int _x=StartIndex();
      
      //'multi-dimension' source data for domain A
      E.Cardinality(2);
                  
      E.Set(0,m_high.GetData(_x+Index)-m_high.GetData(_x+Index+m_lag+m_lag));
      E.Set(1,m_low.GetData(_x+Index)-m_low.GetData(_x+Index+m_lag+m_lag));
   }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void CTrailingCT::SetElement_C(int Index,CElement<double> &E)
   {
      m_high.Refresh(-1);
      m_low.Refresh(-1);
      
      int _x=StartIndex();
      
      //'multi-dimension' source data for domain A
      E.Cardinality(2);
               
      E.Set(0,m_high.GetData(_x+Index)-m_high.GetData(_x+Index+m_lag));
      E.Set(1,m_low.GetData(_x+Index)-m_low.GetData(_x+Index+m_lag));
   }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void CTrailingCT::SetElement_D(int Index,CElement<int> &E)
   {
      //'single-dimension' image data for domain D
      int _value=(20*(Index+1))-120;
      E.Cardinality(1);E.Set(0,_value);
   }
//+------------------------------------------------------------------+
